#include <stdio.h>
#include <stdlib.h>
#include "dict.h"
#include "dict-hash.h"

/* Implemente el iterador para tablas de hashing al final de este archivo */

typedef struct nodo {
  void *llave;
  void *valor;
  struct nodo *prox;
} Nodo;

typedef struct {
  DictOps *ops;
  Comparador comp;
  HashFun f;
  int n;
  Nodo **nodos;
} DictHash;

void definirDictHash(Dict d, void *k, void *v);
int vacioDictHash(Dict d);
void *consultarDictHash(Dict d, void *k);
int obtenerDictHash(Dict d, void *k, void **pk, void **pv);
int eliminarDictHash(Dict d, void *k);
void destruirDictHash(Dict d);

static DictOps dictHashOps= { definirDictHash, vacioDictHash, consultarDictHash,
  obtenerDictHash, eliminarDictHash, destruirDictHash};

Dict nuevoDictHash(int n, Comparador comp, HashFun F) {
  DictHash *d= (DictHash*)malloc(sizeof(DictHash));
  int i;
  d->ops= &dictHashOps;
  d->comp= comp;
  d->n= n;
  d->nodos= (Nodo**)malloc(n * sizeof(Nodo*));
  d->f=F;
  for (i= 0; i<n; i++){
    ((d->nodos)[i])=NULL;
  }
  return (Dict)d;
}

int vacioDictHash(Dict d) {
  DictHash *da= (DictHash*)d;
  int i;
  for (i= 0; i<da->n; i++){
    if ((da->nodos)[i]!=NULL)
      return 0;
  }
  return 1;
}

static Nodo **buscarLlave(DictHash *da, void *k) {
  /* Entrega la direccion del puntero que referencia el nodo cuya llave es k */
  int i= (*da->f)(k) % da->n;
  Nodo **ppnodo= &da->nodos[i];
  while ((*ppnodo)!=NULL) {
    if ((*da->comp)((*ppnodo)->llave, k) == 0)
      return ppnodo;
    ppnodo= &(*ppnodo)->prox;
  }
  return NULL;
}

void definirDictHash(Dict d, void *k, void *v) {
  /* Supone que la llave k no existe previamente en el diccionario */
  DictHash *da= (DictHash*)d;
  int i= (*da->f)(k) % da->n;
  Nodo *pnodoPrev= (da->nodos)[i];
  Nodo *pnodo= (Nodo*)malloc(sizeof(Nodo));
  pnodo->llave= k;
  pnodo->valor= v;
  pnodo->prox= pnodoPrev;    
  da->nodos[i]= pnodo;
}

void *consultarDictHash(Dict d, void *k) {
  DictHash *da= (DictHash*)d;
  Nodo **ppnodo= buscarLlave(da, k);
  return ppnodo==NULL ? NULL : (*ppnodo)->valor;
}

int obtenerDictHash(Dict d, void *k, void **pk, void **pv) {
  DictHash *da= (DictHash*)d;
  Nodo **ppnodo= buscarLlave(da, k);
  if (ppnodo==NULL) {
    return 0;
  }
  else {
    *pk= (*ppnodo)->llave;
    *pv= (*ppnodo)->valor;
    return 1;
  }
}

int eliminarDictHash(Dict d, void *k) {
  DictHash *da=(DictHash*)d;
  Nodo **ppnodo= buscarLlave(da, k);
  Nodo *pnodo;
  if (ppnodo==NULL)
    return 0;

  pnodo= *ppnodo;
  *ppnodo= pnodo->prox;
  free(pnodo);

  return 1;
}

void destruirDictHash(Dict d) {
  DictHash *da= (DictHash*)d;
  free(da->nodos);
  free(da);
}

