#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>

#include "dict.h"
#include "dict-hash.h"
#include "jsocket.h"
#include "util.h"

typedef void *(*Thread_fun)(void *);

Dict global_dict;
pthread_mutex_t global_mtx= PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t global_cond= PTHREAD_COND_INITIALIZER;

/* Funcion de hashing para strings */

int hash_strings(char *s) {
  int i= 0;
  int h= 0;
  while (*s) {
    h+= *s++ << i++;
    if (i>=24)
      i= 0;
  }
  return h;
}

void *serv(int s) {
  char op;
  char *llave;
 
  printf("cliente conectado\n");
  read(s, &op, 1); /* d, q o r */
  llave= getstr(s);

  if (op=='d') {
    char *val= getstr(s);
    pthread_mutex_lock(&global_mtx);
    (*global_dict->ops->definir)(global_dict, llave, val);
    pthread_cond_broadcast(&global_cond);
    pthread_mutex_unlock(&global_mtx);
    sendstr(s, "definicion exitosa");
  }
  else if (op=='q') {
    char *val;
    pthread_mutex_lock(&global_mtx);
    while ((val= global_dict->ops->consultar(global_dict, llave))==NULL)
      pthread_cond_wait(&global_cond, &global_mtx);
    pthread_mutex_unlock(&global_mtx);
    sendstr(s, val);
    free(llave);
  }
  else if (op=='e') {
    int rc;
    char *llave_orig, *val;
    pthread_mutex_lock(&global_mtx);
    rc= global_dict->ops->obtener(global_dict, llave,
                                  (void**)&llave_orig, (void**)&val);
    if (rc) {
      global_dict->ops->eliminar(global_dict, llave);
      free(val); free(llave_orig);
    }
    pthread_mutex_unlock(&global_mtx);
    if (rc)
      sendstr(s, "eliminacion exitosa");
    else
      sendstr(s, "la eliminacion fallo porque no se encontro la llave");
    free(llave);
  }
  else {
    sendstr(s, "operacion invalida");
    free(llave);
  }
  printf("cliente desconectado\n");
  close(s);
  return NULL;
}

int main(int argc, char **argv) {
  long s, s2;
  pthread_t pid;

  global_dict= nuevoDictHash(100, (Comparador)strcmp, (HashFun)hash_strings);
  signal(SIGPIPE, SIG_IGN);
 
  s = j_socket();
 
  if (j_bind(s, 3000) < 0) {
    fprintf(stderr, "bind fallo.  Port 3000 ocupado?\n");
    exit(1);
  }
 
  /* Cada vez que se conecta un cliente le creo un thread */
  for(;;) {
    pthread_attr_t attr;
    s2= j_accept(s);
    pthread_attr_init(&attr);
    if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED) != 0) {
      fprintf(stderr, "No se puede establecer el atributo\n");
    }
    if ( pthread_create(&pid, &attr, (Thread_fun)serv, (void *)s2) != 0) {
      fprintf(stderr, "No pude crear thread para nuevo cliente %ld!!!\n", s2);
      close(s2);
    }
    pthread_attr_destroy(&attr);
  }

  return 0;
}
